from .firestore import FirestoreSink as firestore, \
    ViewModelSink as view_model, FormSink as form, \
    DomainModelSink as domain_model
from .rest import rest_client, rest_server
from .ws import websocket

